import os
import datetime
from finops_crawler import aws

today = datetime.datetime.now().date()
seven_days_ago = today - datetime.timedelta(days=7)

aws_access_key_id = os.getenv('AWS_ACCESS_KEY_ID')
aws_secret_access_key = os.getenv('AWS_SECRET_ACCESS_KEY')
aws_costs_client = aws.costs_api(aws_access_key_id, aws_secret_access_key)
cost_data = aws_costs_client.get_cost(seven_days_ago, today)

print(cost_data)



class CloudAPI:
    def __init__(self, credentials):
        self.credentials = credentials

    def get_cost(self, start_date, end_date):
        raise NotImplementedError("This method should be overridden in subclass")



import os
import datetime
from finops_crawler import openai

today = datetime.datetime.now().date()
seven_days_ago = today - datetime.timedelta(days=7)

org_id = os.getenv('OPENAI_ORG_ID')
api_key = os.getenv('OPENAI_API_KEY')
openai_costs_client = openai.costs_api(org_id, api_key)
cost_data = openai_costs_client.get_cost(seven_days_ago, today)

print(cost_data)



import os
import datetime
from finops_crawler import azure

tenant_id = os.getenv('AZURE_TENANT_ID')
client_id = os.getenv('AZURE_CLIENT_ID')
client_secret = os.getenv('AZURE_CLIENT_SECRET')
azure_costs_client = azure.costs_api(tenant_id, client_id, client_secret)

# Get all subscription ids
subscription_ids = azure_costs_client.get_all_subscriptions()

today = datetime.datetime.now().date()
seven_days_ago = today - datetime.timedelta(days=7)

# Fetch cost and usage data for each subscription
for subscription_id in subscription_ids:
    print(f"Subscription_id: {subscription_id}")

    # cost_data = azure_costs_client.get_cost(subscription_id, seven_days_ago, today)
    cost_data = azure_costs_client.get_cost_detailed(subscription_id, seven_days_ago, today)
    print(cost_data)



import requests
import datetime
import os
from typing import Optional, Union
from finops_crawler.base import CloudAPI

class OpenAIAPI(CloudAPI):
    def __init__(self, openai_org_id: Optional[str] = None, openai_api_key: Optional[str] = None):
        """
            Initialize OpenAIApi instance with OpenAI organization ID and API key.

            Retrieves the OpenAI organization ID and API key from parameters or environment variables if not provided.

            Args:
                openai_org_id (str, optional): OpenAI organization ID. If not provided, the function attempts to get the value from an environment variable named 'OPENAI_ORG_ID'.
                openai_api_key (str, optional): OpenAI API key. If not provided, the function attempts to get the value from an environment variable named 'OPENAI_API_KEY'.

            Raises:
                ValueError: If neither the parameters nor the corresponding environment variables are set.
            
            Note:
                API documentation: https://platform.openai.com/docs/api-reference/authentication
        """
        if openai_org_id is None:
            openai_org_id = os.getenv('OPENAI_ORG_ID')
        if not openai_org_id:
            raise ValueError('OPENAI_ORG_ID not set')

        if openai_api_key is None:
            openai_api_key = os.getenv('OPENAI_API_KEY')
        if not openai_api_key:
            raise ValueError('OPENAI_API_KEY not set')

        self.base_url = 'https://api.openai.com/v1'
        self.headers = {'Authorization': f'Bearer {openai_api_key}', 'OpenAI-Organization': openai_org_id}


    def get_cost(self, start_date: Union[str, datetime.datetime], end_date: Union[str, datetime.datetime]):
        """
        Fetches the daily costs from OpenAI API for a given date range.
        
        The date inputs can either be string (formatted as 'YYYY-MM-DD') or datetime objects.

        Args:
            start_date (Union[str, datetime.datetime]): The start date for the cost data.
            end_date (Union[str, datetime.datetime]): The end date for the cost data.
            
        Returns:
            List[Dict[str, Any]]: A list of daily costs, with each cost represented as a dictionary that includes the timestamp and date.

        Raises:
            ValueError: If the API request is not successful.

        Note:
            The original API response format is enriched with "date" field as a direct transformation from the "timestamp" field. Source: https://community.openai.com/t/how-can-i-check-openai-usage-with-python/117418/10
        """
        if isinstance(start_date, str):
            start_date = datetime.datetime.strptime(start_date, '%Y-%m-%d')
        if isinstance(end_date, str):
            end_date = datetime.datetime.strptime(end_date, '%Y-%m-%d')

        start_date_str = start_date.strftime('%Y-%m-%d')
        end_date_str = end_date.strftime('%Y-%m-%d')
        results_by_time = []

        url = f'{self.base_url}/dashboard/billing/usage?start_date={start_date_str}&end_date={end_date_str}'

        response = requests.get(url, headers=self.headers)

        if response.status_code != 200:
            print(response.reason)
            raise Exception(f'GET /billing/usage failed with status code {response.status_code}.')

        results_by_time = response.json()['daily_costs']

        for item in results_by_time:
            item_date = datetime.datetime.fromtimestamp(item["timestamp"]).strftime('%Y-%m-%d')
            item["date"] = item_date

        return results_by_time



import datetime
import time
import requests
import os
import csv
import io
from typing import Union
from finops_crawler.base import CloudAPI

class AzureAPI(CloudAPI):
    def __init__(self, tenant_id: str, client_id: str, client_secret: str):
        # https://learn.microsoft.com/en-us/azure/active-directory/develop/v2-oauth2-client-creds-grant-flow#first-case-access-token-request-with-a-shared-secret
        if tenant_id is None:
            tenant_id = os.getenv('AZURE_TENANT_ID')
        if not tenant_id:
            raise ValueError('AZURE_TENANT_ID not set')

        if client_id is None:
            client_id = os.getenv('AZURE_CLIENT_ID')
        if not client_id:
            raise ValueError('AZURE_CLIENT_ID not set')

        if client_secret is None:
            client_secret = os.getenv('AZURE_CLIENT_SECRET')
        if not client_secret:
            raise ValueError('AZURE_CLIENT_SECRET not set')

        url = f"https://login.microsoftonline.com/{tenant_id}/oauth2/v2.0/token"

        payload = {
            'grant_type': 'client_credentials',
            'client_id': client_id,
            'client_secret': client_secret,
            'scope': 'https://management.azure.com/.default'
        }
        headers = {'Content-Type': 'application/x-www-form-urlencoded'}
        print("Start auth")
        response = requests.post(url, headers=headers, data=payload, timeout=10)
        response_json = response.json()

        access_token = response_json.get('access_token')
        print("Auth successful")

        self.headers = {'Authorization': f'Bearer {access_token}'}


    def get_all_subscriptions(self):
        # https://azuresdkdocs.blob.core.windows.net/$web/python/azure-mgmt-resource/23.0.0/azure.mgmt.resource.subscriptions.html#module-azure.mgmt.resource.subscriptions
        # make the POST request to the Cost Management API
        url = 'https://management.azure.com/subscriptions?api-version=2020-01-01'

        response = requests.get(url=url, headers=self.headers)
        try:
            response.raise_for_status()
        except requests.HTTPError as e:
            print(f"API returned error {response.status_code}: {response.reason}. Message: {response.json()['error']['message']}")
            raise

        result = response.json()['value']
        if len(result) == 0:
            print("Result retrieved successfully, but it contains no data.")
            raise ValueError("Empty result")

        subscriptions = [s['subscriptionId'] for s in result]
        return subscriptions


    def get_cost(self, subscription_id: str, start_date: Union[str, datetime.datetime], end_date: Union[str, datetime.datetime]):
        if isinstance(start_date, str):
            start_date = datetime.datetime.strptime(start_date, '%Y-%m-%d')
        if isinstance(end_date, str):
            end_date = datetime.datetime.strptime(end_date, '%Y-%m-%d')

        start_date_str = start_date.strftime('%Y-%m-%dT%H:%M:%SZ')
        end_date_str = end_date.strftime('%Y-%m-%dT%H:%M:%SZ')
        print(f"Looking for data between {start_date_str} and {end_date_str}")

        # build the body for the request
        body = {
            'type': 'ActualCost',
            'timeframe': 'Custom',
            'timePeriod': {
            'from': start_date_str,
            'to': end_date_str
            },
            'dataset': {
            'granularity': 'Daily',
            'aggregation': {
                'totalCost': {
                'name': 'Cost',
                'function': 'Sum'
                }
            },
            'grouping': [
                {'type': 'Dimension', 'name': 'ResourceId'},
                {'type': 'Dimension', 'name': 'ChargeType'}
            ]
            }
        }

        data = []

        # make the POST request to the Cost Management API
        url = f'https://management.azure.com/subscriptions/{subscription_id}/providers/Microsoft.CostManagement/query?api-version=2019-11-01'
        response = requests.post(url=url, headers=self.headers, json=body)
        try:
            response.raise_for_status()
        except requests.HTTPError as e:
            print(f"API returned error {response.status_code}: {response.reason}. Message: {response.json()['error']['message']}")
            raise

        result = response.json()['properties']
        if len(result['rows']) > 0:
            column_names = [column['name'] for column in result['columns']]
            rows = [dict(zip(column_names, row)) for row in result['rows']]
        else:
            print("Result retrieved successfully, but it contains no data. It might be a very new subscription.")
            raise ValueError("Empty result")

        print(f"Results in this batch: {len(rows)}")
        data += rows

        # check if there's a nextLink field in the response
        if 'nextLink' in result:
            next_link = result['nextLink']

            # while there is a next link, get the next page of results
            while next_link:
                response = requests.post(next_link, headers=self.headers, json=body)
                if response.status_code == 429:
                    time.sleep(10)
                    continue
                result = response.json()['properties']
                if len(result['rows']) > 0:
                    column_names = [column['name'] for column in result['columns']]
                    rows = [dict(zip(column_names, row)) for row in result['rows']]
                else:
                    print("Result retrieved successfully, but it contains no data. It might be a very new subscription.")
                    raise ValueError("Empty result")

                print(f"Results in this batch: {len(rows)}")
                data += rows

                # update the next_link value
                next_link = result.get('nextLink')
        print(f"Total rows found: {len(data)}")


        if len(data) > 0:
            return data
        else:
            print("Result retreived successfully, but it contains no data. It might be a very new subscription.")
            return None

    def get_cost_detailed(self, subscription_id: str, start_date: Union[str, datetime.datetime], end_date: Union[str, datetime.datetime]):
        # info: https://learn.microsoft.com/en-us/azure/cost-management-billing/automate/automation-ingest-usage-details-overview
        # result: https://learn.microsoft.com/en-us/azure/cost-management-billing/automate/understand-usage-details-fields
        if isinstance(start_date, str):
            start_date = datetime.datetime.strptime(start_date, '%Y-%m-%d')
        if isinstance(end_date, str):
            end_date = datetime.datetime.strptime(end_date, '%Y-%m-%d')

        start_date_str = start_date.strftime('%Y-%m-%d')
        end_date_str = end_date.strftime('%Y-%m-%d')
        print(f"Looking for detailed cost data between {start_date_str} and {end_date_str}")

        body = {
            'metric': 'AmortizedCost',
            'timePeriod': {
                'start': start_date_str,
                'end': end_date_str
            }
        }

        data = []

        # scope here can be many different things, we're using subscriptions to keep the function parameters the same
        # https://learn.microsoft.com/en-us/azure/cost-management-billing/costs/understand-work-scopes#identify-the-resource-id-for-a-scope
        scope = f"subscriptions/{subscription_id}"
        # make the POST request to the Cost Management API
        url = f'https://management.azure.com/{scope}/providers/Microsoft.CostManagement/generateCostDetailsReport?api-version=2022-05-01'
        response = requests.post(url=url, headers=self.headers, json=body)
        try:
            response.raise_for_status()
        except requests.HTTPError as e:
            print(f"API returned error {response.status_code}: {response.reason}. Message: {response.json()['error']['message']}")
            raise

        while response.status_code == 202:
            retry_after = int(response.headers.get('Retry-After', 5))
            url = response.headers.get('Location')
            if url is None:
                raise ValueError("Location for polling missing")
            print(f"Sleeping for {retry_after} seconds before polling again")
            time.sleep(retry_after)
            response = requests.get(url=url, headers=self.headers)
            try:
                response.raise_for_status()
            except requests.HTTPError as e:
                print(f"API returned error {response.status_code}: {response.reason}. Message: {response.json()['error']['message']}")
                raise

        result = response.json()
        if result['status'] == 'Completed':
            manifest = result['manifest']
            print(f"Result contains {manifest['blobCount']} blobs, looping through them")
            for i in range(manifest['blobCount']):
                print(f"Downloading blob: {i}")
                blob_response = requests.get(url=manifest['blobs'][i]['blobLink'])
                csv_file_like_object = io.StringIO(blob_response.text)
                reader = csv.DictReader(csv_file_like_object)
                rows = list(reader)
                data += rows
                print(f"Found {len(rows)} rows in blob {i}")
        else:
            print(f"Result retrieved successfully, but status is {result['status']} instead of Completed")
            raise ValueError("Empty result")

        print(f"Total rows found: {len(data)}")

        if len(data) > 0:
            return data
        else:
            print("Result retreived successfully, but it contains no data. It might be a very new subscription.")
            return None



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['finops_crawler',
 'finops_crawler.aws',
 'finops_crawler.azure',
 'finops_crawler.openai']

package_data = \
{'': ['*']}

install_requires = \
['boto3==1.26.156', 'requests>=2.31.0,<3.0.0']

setup_kwargs = {
    'name': 'finops-crawler',
    'version': '0.1.11',
    'description': "It's a package for importing cost and usage data from various cloud platforms and SaaS tools",
    'long_description': '# finops-crawler\n\nThis project is active. Development is ongoing, and contributions are welcome (see below).\n\n## Basic Usage\n\nThis package is designed to fetch cost/usage data from various platforms (Azure, AWS, OpenAI, and more to follow).\n\n**The main principle is to have as simple an approach as possible along with the absolute minimum set of permissions required**. Thus, the quickstart part of each guide is presented as CLI commands.\n\nThe result, for now, is a Python variable as returned by the API or SDK.\n\n### Platform-specific documentation\n\nEach platform is different and has a separate guide on how to set up an entity with minimum permissions to read the cost data. Specific code snippets to get started are included.\n\n- [Azure](https://github.com/finops-fitness-club/finops-crawler/tree/main/src/finops_crawler/azure)\n- [AWS](https://github.com/finops-fitness-club/finops-crawler/tree/main/src/finops_crawler/aws)\n- [OpenAI](https://github.com/finops-fitness-club/finops-crawler/tree/main/src/finops_crawler/openai)\n\n\n### Overview of the flow\n\n1. Set up a user (technical user/service principal) for querying the data\n2. Set up permissions for that user\n3. Set environment variables to be used by the package\n4. Use the package\n\n### Resulting data\n\nEach platform has its own format. Also, the same date range might produce different results depending on the platform. Some might include the last day, some may not.\n\nThere is an open data specification being developed by FinOps Foundation ([FOCUS](https://focus.finops.org/)). At this point, it\'s still very new and not adopted by the industry. We will keep a close eye on it and support it as soon as feasible.\n\n*Note*: querying long time periods might trigger paginated results. AWS and Azure handle it correctly. OpenAI does not have paginated results as it\'s an undocumented API and it also has not existed yet for a very long time.\n\n### Plans\n\nIncrease breath by expanding to various other tools and platforms (Databricks, GCP, etc.)\n\nIncrease depth by implementing `save_to_postgres` or similar functionality to actually store it.\n\n### Contributing\n\nGot ideas for improvements? We\'d love your input!\n\n1. Fork and clone this repository.\n2. Optionally, choose an issue labeled as "help wanted" or "good first issue".\n3. Submit a pull request with your changes and a clear description.\n\nRemember, all contributions from bug fixes to documentation updates are greatly appreciated!\n\nEnjoy :)\n\nEmail: finops.fitness.club@gmail.com\n',
    'author': 'Lauri Koobas',
    'author_email': 'laurikoobas@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



import datetime
import boto3
from typing import Optional, Union
from botocore.exceptions import BotoCoreError
from finops_crawler.base import CloudAPI

class AWSAPI(CloudAPI):
    def __init__(self, aws_access_key_id: Optional[str] = None, aws_secret_access_key: Optional[str] = None):
        """
        Initialize AWSAPI.

        Args:
            aws_access_key_id (str, optional): AWS Access Key ID. If not provided, boto3 will 
                fall back to the credentials stored in your environment.
            aws_secret_access_key (str, optional): AWS Secret Access Key. If not provided, boto3 
                will fall back to the credentials stored in your environment.
        """
        self.session = boto3.Session(
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
        )

    def get_account_info(self):
        """
        Retrieves the AWS account number for the current user.

        This function uses the AWS Security Token Service (STS) GetCallerIdentity
        operation, which returns details about the IAM user or role whose 
        credentials are used to call the operation.

        Returns:
            str: AWS account number of the current user.

        Note:
            More information about the AWS STS GetCallerIdentity operation can be 
            found in the AWS documentation: 
            https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sts.html#STS.Client.get_caller_identity
        """
        client = self.session.client('sts') 
        identity = client.get_caller_identity()
        return identity['Account']

    def get_all_accounts(self):
        """
        Retrieves a list of all AWS account IDs in the organization.

        This function uses the AWS Organizations ListAccounts operation, which
        returns a list of all the accounts in the organization. If the AWS account
        is not part of any organization, the function will return False.

        Returns:
            list: A list of AWS account IDs in the organization. If the AWS account is 
            not part of any organization, returns False.

        Raises:
            botocore.exceptions.BotoCoreError: If there's an issue when making the request to AWS.

        Note:
            More information about the AWS Organizations ListAccounts operation can be found in the 
            AWS documentation: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/organizations.html#Organizations.Client.list_accounts
        """
        client = self.session.client('organizations') 
        paginator = client.get_paginator('list_accounts')
        accounts = []
        try:
            for page in paginator.paginate():
                for account in page['Accounts']:
                    accounts.append(account['Id'])
                    print(account['Id'])
        except client.exceptions.AWSOrganizationsNotInUseException as e:
            print("Your account is not a member of an organization.")
            return False

        except BotoCoreError as e:
            print(f"Unexpected AWS error: {e}")

        return accounts

    def get_cost(self, start_date: Union[str, datetime.datetime], end_date: Union[str, datetime.datetime]):
        """
        Retrieves the cost of AWS services used over a specified time period.

        This function calls the AWS Cost Explorer GetCostAndUsage operation, which
        provides metrics associated with your AWS costs, split by service and usage type.
        Costs are presented on a daily basis.

        Args:
            start_date (datetime.datetime): The start date for retrieving AWS cost data.
                Data is returned for a time period that begins at the start date and
                ends at the end date.
            end_date (datetime.datetime): The end date for retrieving AWS cost data.
                Data is returned for a time period that begins at the start date and
                ends at the end date.

        Returns:
            list: A list of results by time. Each result includes the time period and metrics
            associated with the AWS costs, grouped by service and usage type.

        Raises:
            botocore.exceptions.BotoCoreError: If there's an issue when making the request to AWS.
            botocore.exceptions.ClientError: If there's a client-side error with boto3.

        Note:
            More information about the AWS Cost Explorer GetCostAndUsage operation can be found in the 
            AWS documentation: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ce.html#CostExplorer.Client.get_cost_and_usage
        """
        if isinstance(start_date, str):
            start_date = datetime.datetime.strptime(start_date, '%Y-%m-%d')
        if isinstance(end_date, str):
            end_date = datetime.datetime.strptime(end_date, '%Y-%m-%d')

        start_date_str = start_date.strftime('%Y-%m-%d')
        end_date_str = end_date.strftime('%Y-%m-%d')
        results_by_time = []

        client = self.session.client('ce')

        response = client.get_cost_and_usage(
            TimePeriod={
                'Start': start_date_str,
                'End': end_date_str
            },
            Granularity='DAILY',
            Metrics=[
                'UnblendedCost',
            ],
            GroupBy=[
                {'Type': 'DIMENSION', 'Key': 'SERVICE'},
                {'Type': 'DIMENSION', 'Key': 'USAGE_TYPE'},
            ],
        )
        results_by_time += response['ResultsByTime']
        if 'NextPageToken' in response:
            next_page_token = response['NextPageToken']
            while True:
                response = client.get_cost_and_usage(
                    TimePeriod={
                        'Start': start_date_str,
                        'End': end_date_str
                    },
                    Granularity='DAILY',
                    Metrics=[
                        'UnblendedCost',
                    ],
                    GroupBy=[
                        {'Type': 'DIMENSION', 'Key': 'SERVICE'},
                        {'Type': 'DIMENSION', 'Key': 'USAGE_TYPE'},
                    ],
                    NextPageToken = next_page_token
                )
                results_by_time += response['ResultsByTime']
                if 'NextPageToken' in response:
                    next_page_token = response['NextPageToken']
                else:
                    break

        return results_by_time
